// ------------------------------------------------------------------
// test-governor - programme de test des heuristiques pour le sous-
//                 systeme cpufreq du noyau Linux.
//
// (c) 2011 Christophe BLAESS - Tous droits reserves.
//
// Exemple pour un post du blog http://www.blaess.fr/christophe/
// ------------------------------------------------------------------

#define _GNU_SOURCE // Pour avoir pthread_attr_setaffinity_np()

#include <errno.h>
#include <fcntl.h>
#include <pthread.h>
#include <sched.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <sys/time.h>

// Parametrage du programme :
//   Le CPU (ou coeur) sur lequel le thread va s'excuter
#    define CPU_THREAD               1
//   La priorite temps-reel du thread (0 = pas temps-reel)
#    define PRIORITE_THREAD          10
//   Le nombre d'iteration dans une boucle active
#    define NB_ITERATIONS_PAR_BOUCLE 200000
//   Le nombre de boucles actives a executer
#    define NB_BOUCLES               100


static void * thread (void * arg);

int main(int argc, char * argv[])
{
	int fd;
	char filename[256];
	cpu_set_t  cpuset;
	pthread_t thr;
	pthread_attr_t attr;
	struct sched_param param;

	if (argc != 2) {
		fprintf(stderr, "usage: %s <governor>\n", argv[0]);
		exit(1);
	}

	// Basculer l'heuristique de cpufreq
	sprintf(filename, "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_governor", CPU_THREAD);

	fd = open(filename, O_RDWR);
	if (fd < 0) {
		perror(filename);
		exit(2);
	}
	if (write(fd, argv[1], strlen(argv[1])) < 0) {
		perror(argv[1]);
		exit(2);
	}
	close (fd);

	// Preparer le demarrage du thread sur son CPU
	pthread_attr_init(& attr);

	CPU_ZERO(& cpuset);
	CPU_SET(CPU_THREAD, & cpuset);
	if (pthread_attr_setaffinity_np(& attr, sizeof(cpuset), & cpuset) != 0) {
		perror("pthread_attr_setaffinity_np");
		exit(2);
	}

	// Avec son ordonnancement temps-reel
	if (PRIORITE_THREAD > 0) {
		if (pthread_attr_setschedpolicy(&attr, SCHED_FIFO) != 0) {
			perror("pthread_attr_setschedpolicy");
			exit(2);
		}
		param.sched_priority = PRIORITE_THREAD;
		if (pthread_attr_setschedparam(&attr, &param) != 0) {
			perror("pthread_attr_setschedparam");
			exit(2);
		}
		// sans tenir compte de l'ordonnancement du thread main()
		if (pthread_attr_setinheritsched(& attr, PTHREAD_EXPLICIT_SCHED) != 0) {
			perror("pthread_attr_setinheritsched");
			exit(2);
		}
	}
	// Lancer le thread
	if ((errno = pthread_create(& thr, &attr, thread, NULL)) != 0) {
		perror("pthread_create");
		exit(2);
	}
	// Attendre qu'il se termine
	pthread_join(thr, NULL);

	return EXIT_SUCCESS;
}



static struct timeval tv[NB_BOUCLES+1];

static void * thread(void * unused)
{
	int i;
	int j;
	long long int duree;

	// Petite pause de stabilisation
	sleep(1);

	// Puis les boucles actives
	for (i = 0; i < NB_BOUCLES; i ++) {
		gettimeofday(& (tv[i]), NULL);
		for (j = 0; j < NB_ITERATIONS_PAR_BOUCLE; j ++)
			;
	}
	gettimeofday(& (tv[i]), NULL);

	// Affichage des durees des boucles
	for (i = 0; i < NB_BOUCLES; i ++) {
		duree  = tv[i+1].tv_sec - tv[i].tv_sec;
		duree *= 1000000;
		duree += tv[i+1].tv_usec - tv[i].tv_usec;
		fprintf(stdout, "%4d %6lld\n", i, duree);
	}
	pthread_exit(0);
}



